"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.UtxoFullTransaction = void 0;
const types_1 = require("../../../types");
const errors_1 = require("../../../utils/errors");
const utils_1 = require("../../../utils/utils");
const utxoUtils_1 = require("../../../utils/utxoUtils");
const TxBase_1 = require("./TxBase");
function isCoinbaseVin(vin) {
    return vin.coinbase !== undefined;
}
class UtxoFullTransaction extends TxBase_1.TxBaseFull {
    get type() {
        if (this.data.vin.length === 0 || isCoinbaseVin(this.data.vin[0])) {
            return "coinbase";
        }
        else
            return "payment";
    }
    get isNativePayment() {
        // On these chains there are no other types of transactions
        return true;
    }
    get txid() {
        return this.data.txid;
    }
    get stdTxid() {
        return (0, utils_1.unPrefix0x)(this.txid);
    }
    get reference() {
        const references = [];
        for (const vo of this.data.vout) {
            if (vo.scriptPubKey.hex.substring(0, 2) == (0, utils_1.unPrefix0x)((0, utils_1.toHex)(utxoUtils_1.WordToOpcode.OP_RETURN))) {
                const dataSplit = vo.scriptPubKey.asm.split(" ");
                references.push(dataSplit[1]);
            }
        }
        return references;
    }
    get stdPaymentReference() {
        let paymentReference = this.reference.length === 1 ? (0, utils_1.prefix0x)(this.reference[0]) : "";
        if (!(0, utils_1.isValidBytes32Hex)(paymentReference)) {
            paymentReference = utils_1.ZERO_BYTES_32;
        }
        return paymentReference;
    }
    get unixTimestamp() {
        return this.data.blocktime;
    }
    get successStatus() {
        return types_1.TransactionSuccessStatus.SUCCESS;
    }
    get sourceAddresses() {
        return this.data.vin.map((vin) => {
            if (isCoinbaseVin(vin)) {
                return undefined;
            }
            else {
                return vin.prevout.scriptPubKey.address; // are we sure that every prevout has an address
            }
        });
    }
    toBnValue(value) {
        if (value === undefined) {
            return (0, utils_1.toBN)(0);
        }
        return (0, utils_1.toBN)(Math.round(value * this.elementaryUnits.toNumber()).toFixed(0));
    }
    get spentAmounts() {
        return this.data.vin.map((vin) => {
            if (isCoinbaseVin(vin)) {
                return {
                    amount: (0, utils_1.toBN)(0),
                };
            }
            else {
                const address = vin.prevout.scriptPubKey.address; // are we sure that every prevout has an address
                const value = this.toBnValue(vin.prevout.value || 0);
                const addressAmount = {
                    address: address,
                    amount: value,
                    utxo: vin.vout || 0,
                };
                return addressAmount;
            }
        });
    }
    get intendedSpentAmounts() {
        return this.spentAmounts;
    }
    get receivingAddresses() {
        /* istanbul ignore if */
        // Transaction should always have vout array
        if (!this.data.vout) {
            return [];
        }
        return this.data.vout.map((vout) => vout.scriptPubKey.address);
    }
    get receivedAmounts() {
        return this.data.vout.map((vout) => {
            return {
                address: vout.scriptPubKey.address,
                amount: this.toBnValue(vout.value),
                utxo: vout.n,
            };
        });
    }
    get intendedReceivedAmounts() {
        return this.receivedAmounts;
    }
    ///////////////////////////////
    //// Utxo specific methods ////
    ///////////////////////////////
    /**
     * Asserts whether the vin index is in valid range. If not, exception is thrown.
     * @param vinIndex vin index
     */
    assertValidVinIndex(vinIndex) {
        if (vinIndex < 0 || vinIndex >= this.sourceAddresses.length) {
            throw new errors_1.mccError(errors_1.mccErrorCode.InvalidParameter, Error("Invalid vin index"));
        }
    }
    /**
     * Asserts whether the vout index is in valid range. If not, exception is thrown.
     * @param voutIndex vout index
     */
    assertValidVoutIndex(voutIndex) {
        if (voutIndex < 0 || voutIndex >= this.receivingAddresses.length) {
            throw new errors_1.mccError(errors_1.mccErrorCode.InvalidParameter, Error("Invalid vout index"));
        }
    }
    paymentSummary({ inUtxo, outUtxo }) {
        try {
            this.assertValidVinIndex(inUtxo);
        }
        catch (e) {
            return { status: TxBase_1.PaymentSummaryStatus.InvalidInUtxo };
        }
        try {
            this.assertValidVoutIndex(outUtxo);
        }
        catch (e) {
            return { status: TxBase_1.PaymentSummaryStatus.InvalidOutUtxo };
        }
        if (this.type === "coinbase") {
            return { status: TxBase_1.PaymentSummaryStatus.Coinbase };
        }
        const spentAmount = this.spentAmounts[inUtxo];
        if (!spentAmount.address) {
            return { status: TxBase_1.PaymentSummaryStatus.NoSpentAmountAddress };
        }
        const receiveAmount = this.receivedAmounts[outUtxo];
        if (!receiveAmount.address) {
            return { status: TxBase_1.PaymentSummaryStatus.NoReceiveAmountAddress };
        }
        // Extract addresses from input and output fields
        const sourceAddress = spentAmount.address;
        const receivingAddress = receiveAmount.address;
        // We will update this once we iterate over inputs and outputs if we have full transaction
        let oneToOne = true;
        let inFunds = (0, utils_1.toBN)(0);
        let returnFunds = (0, utils_1.toBN)(0);
        let outFunds = (0, utils_1.toBN)(0);
        let inFundsOfReceivingAddress = (0, utils_1.toBN)(0);
        for (const vinAmount of this.spentAmounts) {
            if (sourceAddress && vinAmount.address === sourceAddress) {
                inFunds = inFunds.add(vinAmount.amount);
            }
            if (receivingAddress && vinAmount.address === receivingAddress) {
                inFundsOfReceivingAddress = inFundsOfReceivingAddress.add(vinAmount.amount);
            }
            if (oneToOne && vinAmount.address != sourceAddress && vinAmount.address != receivingAddress) {
                oneToOne = false;
            }
        }
        for (const voutAmount of this.receivedAmounts) {
            if (sourceAddress && voutAmount.address === sourceAddress) {
                returnFunds = returnFunds.add(voutAmount.amount);
            }
            if (receivingAddress && voutAmount.address === receivingAddress) {
                outFunds = outFunds.add(voutAmount.amount);
            }
            // Outputs without address do not break one-to-one condition
            if (oneToOne && !voutAmount.address && voutAmount.amount.gt((0, utils_1.toBN)(0))) {
                oneToOne = false;
            }
            if (oneToOne && voutAmount.address && voutAmount.address != sourceAddress && voutAmount.address != receivingAddress) {
                oneToOne = false;
            }
        }
        return {
            status: TxBase_1.PaymentSummaryStatus.Success,
            response: {
                blockTimestamp: this.unixTimestamp,
                transactionHash: this.stdTxid,
                sourceAddressHash: (0, utils_1.standardAddressHash)(sourceAddress),
                sourceAddress,
                receivingAddress,
                receivingAddressHash: (0, utils_1.standardAddressHash)(receivingAddress),
                spentAmount: inFunds.sub(returnFunds),
                receivedAmount: outFunds.sub(inFundsOfReceivingAddress),
                transactionStatus: this.successStatus,
                paymentReference: this.stdPaymentReference,
                intendedSourceAmount: inFunds.sub(returnFunds),
                intendedReceivingAddressHash: (0, utils_1.standardAddressHash)(receivingAddress),
                intendedReceivingAddress: receivingAddress,
                intendedReceivingAmount: outFunds.sub(inFundsOfReceivingAddress),
                oneToOne,
            },
        };
    }
    balanceDecreasingSummary({ sourceAddressIndicator }) {
        // We expect sourceAddressIndicator to be utxo vin index (as hex string)
        if (!(0, utils_1.isValidBytes32Hex)(sourceAddressIndicator)) {
            return { status: TxBase_1.BalanceDecreasingSummaryStatus.NotValidSourceAddressFormat };
        }
        const vinIndex = parseInt(sourceAddressIndicator, 16);
        if (isNaN(vinIndex)) {
            return { status: TxBase_1.BalanceDecreasingSummaryStatus.NotValidSourceAddressFormat };
        }
        try {
            this.assertValidVinIndex(vinIndex);
        }
        catch (e) {
            return { status: TxBase_1.BalanceDecreasingSummaryStatus.InvalidInUtxo };
        }
        const transactionType = this.type;
        switch (transactionType) {
            case "coinbase": {
                return { status: TxBase_1.BalanceDecreasingSummaryStatus.Coinbase };
            }
            case "payment": {
                const spentAmounts = this.spentAmounts;
                const spentAmount = spentAmounts[vinIndex];
                const sourceAddress = spentAmount.address;
                if (sourceAddress) {
                    let inFunds = (0, utils_1.toBN)(0);
                    let returnFunds = (0, utils_1.toBN)(0);
                    for (const vinAmount of this.spentAmounts) {
                        if (vinAmount.address === sourceAddress) {
                            inFunds = inFunds.add(vinAmount.amount);
                        }
                    }
                    for (const voutAmount of this.receivedAmounts) {
                        if (voutAmount.address === sourceAddress) {
                            returnFunds = returnFunds.add(voutAmount.amount);
                        }
                    }
                    return {
                        status: TxBase_1.BalanceDecreasingSummaryStatus.Success,
                        response: {
                            blockTimestamp: this.unixTimestamp,
                            transactionHash: this.stdTxid,
                            sourceAddressIndicator: sourceAddressIndicator,
                            sourceAddressHash: (0, utils_1.standardAddressHash)(sourceAddress),
                            sourceAddress: sourceAddress,
                            spentAmount: inFunds.sub(returnFunds),
                            transactionStatus: this.successStatus,
                            paymentReference: this.stdPaymentReference,
                            isFull: true,
                        },
                    };
                }
                else {
                    // The input has no address, the type is wrongly extracted (Should not happen)
                    return { status: TxBase_1.BalanceDecreasingSummaryStatus.InvalidTransactionDataObject };
                }
                return { status: TxBase_1.BalanceDecreasingSummaryStatus.Success }; // TODO: implement
            }
            default:
                // exhaustive switch guard: if a compile time error appears here, you have forgotten one of the cases
                // eslint-disable-next-line @typescript-eslint/no-empty-function, @typescript-eslint/no-unused-vars
                ((_) => { })(transactionType);
        }
        // We didn't find the address we are looking for
        return { status: TxBase_1.BalanceDecreasingSummaryStatus.NoSourceAddress };
    }
    get hash() {
        throw new Error("Method not implemented.");
    }
}
exports.UtxoFullTransaction = UtxoFullTransaction;
//# sourceMappingURL=UtxoTrancation.js.map